@@ -1,18 +1,22 @@
+import sys
+
 try:
     from paste.registry import StackedObjectProxy
     beaker_session = StackedObjectProxy()
 except:
     pass
 from myghtyutils.session import Session
+from myghtyutils.util import PrefixArgs
 
 
 class request(object):
-    class headers_out(object):
-        def __init__(self):
-            self.headers = {}
-        def add(self, name, value):
-            self.headers[name] = value
-    headers_out = headers_out()
+    def __init__(self):
+        class headers_out(object):
+            def __init__(self):
+                self.headers = {}
+            def add(self, name, value):
+                self.headers[name] = value
+        self.headers_out = headers_out()
 
 class SessionObject(object):
     """Session proxy/lazy creator
@@ -26,20 +30,18 @@ class SessionObject(object):
     def __init__(self, environ, **params):
         self.__dict__['_params'] = params
         self.__dict__['_environ'] = environ
-        self.__dict__['_sessinit'] = False
         self.__dict__['_sess'] = None
         self.__dict__['_headers'] = []
     
     def _session(self):
         """Lazy initial creation of session object"""
-        if not self.__dict__['_sessinit']:
+        if not self.__dict__['_sess']:
             params = self.__dict__['_params']
             environ = self.__dict__['_environ']
             req = request()
             self.__dict__['_headers'] = req.headers_out
             req.headers_in = dict(cookie=environ.get('HTTP_COOKIE'))
             self.__dict__['_sess'] = Session(req, use_cookies=True, **params)
-            self.__dict__['_sessinit'] = True
         return self.__dict__['_sess']
     
     def __getattr__(self, attr):
@@ -76,31 +78,36 @@ session_params = dict(invalidate_corrupt = False, type = None, data_dir = None,
 
 class SessionMiddleware(object):
     def __init__(self, wrap_app, global_conf={}, auto_register=True, **params):
-        instance_params = session_params.copy()
-        for key in global_conf:
-            if key.startswith('session_'):
-                if key[8:] in instance_params:
-                    instance_params[key[8:]] = global_conf[key]
-        
-        instance_params.update(params)
+        self.params = SessionArgs(**session_params)
+        self.params.set_prefix_params(**global_conf)
+        self.params.set_prefix_params(**params)
         self.auto_register = auto_register
-        self.params = instance_params
-        self.key = self.params['key']
         self.wrap_app = wrap_app
     
     def __call__(self, environ, start_response):
-        session = SessionObject(environ, **self.params)
+        session = SessionObject(environ, **self.params.params)
         if environ.get('paste.registry') and self.auto_register:
             environ['paste.registry'].register(beaker_session, session)
         environ['beaker.session'] = session
         
         def session_start_response(status, headers, exc_info = None):
-            if session.__dict__['_sessinit']:
+            if session.__dict__['_sess']:
                 cookie = session.__dict__['_headers'].headers.get('set-cookie')
                 if cookie:
                     headers.append(('Set-cookie', cookie))
             return start_response(status, headers, exc_info)
-        return self.wrap_app(environ, session_start_response)
+        try:
+            response = self.wrap_app(environ, session_start_response)
+        except:
+            ty, val = sys.exc_info()[:2]
+            if isinstance(ty, str):
+                raise ty, val, sys.exc_info()[2]
+            if ty.__name__ == 'HTTPFound' and session.__dict__['_sess']:
+                cookie = session.__dict__['_headers'].headers.get('set-cookie')
+                val.headers.append(('Set-cookie', cookie))
+            raise ty, val, sys.exc_info()[2]
+        else:
+            return response
 
 def session_filter_factory(global_conf, **kwargs):
     def filter(app):
@@ -109,3 +116,14 @@ def session_filter_factory(global_conf, **kwargs):
 
 def session_filter_app_factory(app, global_conf, **kwargs):
     return SessionMiddleware(app, global_conf, **kwargs)
+
+class SessionArgs(PrefixArgs):
+    def __init__(self, **params):
+        PrefixArgs.__init__(self, 'session_')
+        self.set_prefix_params(**params)
+        
+    def clone(self, **params):
+        p = self.get_params(**params)
+        arg = CacheArgs()
+        arg.params = p
+        return arg
